import contextlib
import io
import os
import typing

import boto3
from botocore.exceptions import ClientError
from pyjackson.decorators import cached_property

from algolink.config import Config, Core, Param
from algolink.core.errors import ArtifactExistsError, NoSuchArtifactError
from algolink.core.objects.artifacts import ArtifactCollection, Blob, Blobs, StreamContextManager
from algolink.repository.artifact import ArtifactRepository
from algolink.utils.log import logger


class S3Config(Config):
    namespace = 's3'
    ACCESS_KEY = Param('access_key')
    SECRET_KEY = Param('secret_key')


if Core.DEBUG:
    S3Config.log_params()


class _WithS3Client:
    def __init__(self, bucket_name: str, endpoint: str = None, region: str = None):
        self.bucket_name = bucket_name
        self.endpoint = endpoint
        self.region = region

    @cached_property
    def _s3(self):
        logger.debug('Creating s3 client with endpoint %s', self.endpoint)
        return boto3.client('s3',
                            endpoint_url=self.endpoint,
                            aws_access_key_id=S3Config.ACCESS_KEY,
                            aws_secret_access_key=S3Config.SECRET_KEY,
                            region_name=self.region)

    @cached_property
    def _s3_res(self):
        logger.debug('Creating s3 resource with endpoint %s', self.endpoint)
        return boto3.resource('s3',
                              endpoint_url=self.endpoint,
                              aws_access_key_id=S3Config.ACCESS_KEY,
                              aws_secret_access_key=S3Config.SECRET_KEY,
                              region_name=self.region)


class S3Blob(Blob, _WithS3Client):
    """
    :class:`.Blob` implementation which stores artifacts in Amazon S3-compatible file system

    S3 credentials are to be specified through `S3_ACCESS_KEY` and `S3_SECRET_KEY` environment variables.

    :param: s3path: S3 path to the artifact represented by this object
    :param: bucket_name: name of S3 bucket to use for storage
    :param: endpoint: HTTP URL of S3 server to connect to
    """
    type = 's3'

    def __init__(self, s3path: str, bucket_name: str, endpoint: str = None):
        _WithS3Client.__init__(self, bucket_name, endpoint)
        self.s3path = s3path

    def materialize(self, path):
        logger.debug('Downloading file from %s to %s', self.s3path, path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        self._s3.download_file(self.bucket_name, self.s3path, path)

    @contextlib.contextmanager
    def bytestream(self) -> StreamContextManager:
        logger.debug('Streaming file from %s', self.s3path)
        yield self._s3.get_object(Bucket=self.bucket_name, Key=self.s3path)['Body']


class S3ArtifactRepository(ArtifactRepository, _WithS3Client):
    """
    :class:`.ArtifactRepository` implementation which stores artifacts in Amazon S3-compatible file system

    S3 credentials are to be specified through `S3_ACCESS_KEY` and `S3_SECRET_KEY` environment variables.

    :param: bucket_name: name of S3 bucket to use for storage
    :param: endpoint: HTTP URL of S3 server to connect to
    """

    type = 's3'

    def _ensure_bucket(self):
        if not self._bucket_exists():
            self._s3.create_bucket(Bucket=self.bucket_name)

    def _bucket_exists(self):
        try:
            self._s3.head_bucket(Bucket=self.bucket_name)
            return True
        except ClientError:
            return False

    def _list_blobs(self, prefix):
        bs = self._s3.list_objects(Bucket=self.bucket_name, Prefix=prefix)
        if 'Contents' not in bs:
            return {}
        return {o['Key']: o for o in bs['Contents']}

    def get_artifact(self, artifact_type, artifact_id: str) -> ArtifactCollection:
        artifact_id = f'{artifact_type}/{artifact_id}'
        if not self._bucket_exists():
            raise NoSuchArtifactError(artifact_id, self)

        keys = list(self._list_blobs(artifact_id).keys())
        if len(keys) == 0:
            raise NoSuchArtifactError(artifact_id, self)
        elif len(keys) == 1 and keys[0] == '.':
            return Blobs({})
        else:
            return Blobs({
                os.path.relpath(key, artifact_id): S3Blob(key, self.bucket_name, self.endpoint)
                for key in keys
            })

    def push_artifact(self, artifact_type, artifact_id: str, blobs: typing.Dict[str, Blob]) -> ArtifactCollection:
        artifact_id = f'{artifact_type}/{artifact_id}'
        self._ensure_bucket()

        if len(self._list_blobs(artifact_id)) > 0:
            raise ArtifactExistsError(artifact_id, self)

        if len(blobs) == 0:
            self._s3.upload_fileobj(io.BytesIO(b''), self.bucket_name, artifact_id)
            return Blobs({})

        result = {}
        for filepath, blob in blobs.items():
            join = os.path.join(artifact_id, filepath)
            with blob.bytestream() as b:
                logger.debug('Uploading %s to s3 %s/%s', blob, self.endpoint, self.bucket_name)
                self._s3.upload_fileobj(b, self.bucket_name, join)
            result[filepath] = S3Blob(join, self.bucket_name, self.endpoint)
        return Blobs(result)

    def delete_artifact(self, artifact_type, artifact_id: str):
        artifact_id = f'{artifact_type}/{artifact_id}'
        if not self._bucket_exists():
            raise NoSuchArtifactError(artifact_id, self)
        keys = list(self._list_blobs(artifact_id).keys())
        if len(keys) == 0:
            raise NoSuchArtifactError(artifact_id, self)
        else:
            logger.debug('Deleting %s from %s/%s', artifact_id, self.endpoint, self.bucket_name)
            self._s3.delete_objects(Bucket=self.bucket_name, Delete={'Objects': [{'Key': k} for k in keys]})



import copy
import os
from typing import Dict, List, Optional, Set, Tuple, Union

import pyjackson

from algolink.core.errors import (EnvironmentWithInstancesError, ExistingEnvironmentError, ExistingImageError,
                                 ExistingInstanceError, ExistingModelError, ExistingPipelineError, ExistingProjectError,
                                 ExistingTaskError, ImageWithInstancesError, NonExistingEnvironmentError,
                                 NonExistingImageError, NonExistingInstanceError, NonExistingModelError,
                                 NonExistingPipelineError, NonExistingProjectError, NonExistingTaskError,
                                 ProjectWithTasksError, TaskWithFKError)
from algolink.core.objects.core import Image, Model, Pipeline, Project, RuntimeEnvironment, RuntimeInstance, Task
from algolink.repository.metadata.base import MetadataRepository, ProjectVar, TaskVar, bind_to_self
from algolink.utils.log import logger

_Projects = Dict[int, Project]
_Tasks = Dict[int, Task]
_Models = Dict[int, Model]
_Pipelines = Dict[int, Pipeline]
_Images = Dict[int, Image]
_Environments = Dict[int, RuntimeEnvironment]
_Instances = Dict[int, RuntimeInstance]


class _LocalContainer:
    def __init__(self, next_project_id: int = 0, projects: _Projects = None,
                 next_task_id: int = 0, tasks: _Tasks = None,
                 next_model_id: int = 0, models: _Models = None,
                 next_pipeline_id: int = 0, pipelines: _Pipelines = None,
                 next_image_id: int = 0, images: _Images = None,
                 next_environment_id: int = 0, environments: _Environments = None,
                 next_instance_id: int = 0, instances: _Instances = None):
        self.next_project_id = next_project_id
        self.projects: _Projects = {}
        self.project_name_index: Dict[str, int] = {}

        self.next_task_id = next_task_id
        self.tasks: _Tasks = {}
        self.task_name_index: Dict[Tuple[int, str], int] = {}

        self.next_model_id = next_model_id
        self.models: _Models = {}
        self.model_name_index: Dict[Tuple[int, str], int] = {}

        self.next_pipeline_id = next_pipeline_id
        self.pipelines: _Pipelines = {}
        self.pipeline_name_index: Dict[Tuple[int, str], int] = {}

        self.next_image_id = next_image_id
        self.images: _Images = {}
        self.image_name_index: Dict[Tuple[int, str], int] = {}

        self.next_environment_id = next_environment_id
        self.environments: _Environments = {}
        self.environment_name_index: Dict[str, int] = {}

        self.next_instance_id = next_instance_id
        self.instances: _Instances = {}
        self.instance_name_index: Dict[Tuple[int, int, str], int] = {}
        self.instance_index: Dict[Tuple[int, int], Set[int]] = {}
        self.image_instance: Dict[int, Set[int]] = {}
        self.environment_instance: Dict[int, Set[int]] = {}

        for p in (projects or {}).values():
            self.add_project(p)

        for t in (tasks or {}).values():
            self.add_task(t)

        for m in (models or {}).values():
            self.add_model(m)

        for p in (pipelines or {}).values():
            self.add_pipeline(p)

        for i in (images or {}).values():
            self.add_image(i)

        for e in (environments or {}).values():
            self.add_environment(e)

        for i in (instances or {}).values():
            self.add_instance(i)

    def get_and_increment(self, name):
        next_id = getattr(self, name)
        setattr(self, name, next_id + 1)
        return next_id

    def add_project(self, project: Project):
        assert project.id is not None
        self.projects[project.id] = project
        self.project_name_index[project.name] = project.id

    def get_project_by_id(self, project_id):
        return self.projects.get(project_id)

    def get_project_by_name(self, name: str):
        return self.get_project_by_id(self.project_name_index.get(name, None))

    def remove_project(self, project_id):
        project = self.projects.pop(project_id, None)
        del self.project_name_index[project.name]
        return project

    def add_task(self, task: Task):
        assert task.id is not None
        self.tasks[task.id] = task
        self.task_name_index[(task.project_id, task.name)] = task.id
        self.projects[task.project_id]._tasks.add(task)

    def get_task_by_id(self, task_id):
        return self.tasks.get(task_id)

    def get_task_by_name(self, project_id: int, name: str):
        return self.get_task_by_id(self.task_name_index.get((project_id, name), None))

    def remove_task(self, task_id):
        task = self.tasks.pop(task_id, None)

        self.task_name_index.pop((task.project_id, task.name), None)
        del self.projects[task.project_id]._tasks[task.id]
        return task

    def add_model(self, model: Model):
        assert model.id is not None
        self.models[model.id] = model
        self.model_name_index[(model.task_id, model.name)] = model.id
        self.tasks[model.task_id]._models.add(model)

    def get_model_by_id(self, model_id):
        return self.models.get(model_id, None)

    def get_model_by_name(self, task_id: int, name: str):
        return self.get_model_by_id(self.model_name_index.get((task_id, name), None))

    def remove_model(self, model_id):
        model = self.models.pop(model_id, None)
        self.model_name_index.pop((model.task_id, model.name), None)
        del self.tasks[model.task_id]._models[model.id]
        return model

    def add_pipeline(self, pipeline: Pipeline):
        assert pipeline.id is not None
        self.pipelines[pipeline.id] = pipeline
        self.pipeline_name_index[(pipeline.task_id, pipeline.name)] = pipeline.id
        self.tasks[pipeline.task_id]._pipelines.add(pipeline)

    def get_pipeline_by_id(self, pipeline_id):
        return self.pipelines.get(pipeline_id, None)

    def get_pipeline_by_name(self, task_id: int, name: str):
        return self.get_pipeline_by_id(self.pipeline_name_index.get((task_id, name), None))

    def remove_pipeline(self, pipeline_id):
        pipeline = self.pipelines.pop(pipeline_id, None)
        self.pipeline_name_index.pop((pipeline.task_id, pipeline.name), None)
        del self.tasks[pipeline.task_id]._pipelines[pipeline.id]
        return pipeline

    def add_image(self, image: Image):
        assert image.id is not None
        self.images[image.id] = image
        self.image_name_index[(image.task_id, image.name)] = image.id
        self.tasks[image.task_id]._images.add(image)

    def get_image_by_id(self, image_id):
        return self.images.get(image_id, None)

    def get_image_by_name(self, task_id: int, name: str):
        return self.get_image_by_id(self.image_name_index.get((task_id, name), None))

    def remove_image(self, image_id):
        image = self.images.pop(image_id, None)
        self.image_name_index.pop((image.task_id, image.name), None)
        del self.tasks[image.task_id]._images[image.id]
        return image

    def add_environment(self, environment: RuntimeEnvironment):
        assert environment.id is not None
        self.environments[environment.id] = environment
        self.environment_name_index[environment.name] = environment.id

    def get_environment_by_id(self, environment_id):
        return self.environments.get(environment_id)

    def get_environment_by_name(self, name: str):
        return self.get_environment_by_id(self.environment_name_index.get(name, None))

    def remove_environment(self, environment_id):
        environment = self.environments.pop(environment_id, None)
        del self.environment_name_index[environment.name]
        return environment

    def add_instance(self, instance: RuntimeInstance):
        assert instance.id is not None
        self.instances[instance.id] = instance
        self.instance_name_index[(instance.environment_id, instance.image_id, instance.name)] = instance.id
        self.instance_index.setdefault((instance.environment_id, instance.image_id), set()).add(instance.id)
        self.environment_instance.setdefault(instance.environment_id, set()).add(instance.id)
        self.image_instance.setdefault(instance.image_id, set()).add(instance.id)

    def get_instance_by_id(self, instance_id: int):
        return self.instances.get(instance_id, None)

    def get_instance_by_name(self, environment_id: int, image_id: int, name: str):
        return self.get_instance_by_id(self.instance_name_index.get((environment_id, image_id, name), None))

    def get_instances(self, environment_id: int, image_id: int):
        return [self.get_instance_by_id(iid) for iid in self.instance_index.get((environment_id, image_id), set())]

    def get_instances_by_image_id(self, image_id):
        return [self.get_instance_by_id(iid) for iid in self.image_instance.get(image_id, set())]

    def get_instances_by_environment_id(self, environment_id):
        return [self.get_instance_by_id(iid) for iid in self.environment_instance.get(environment_id, set())]

    def remove_instance(self, instance_id: int):
        instance = self.instances.pop(instance_id, None)
        self.instance_name_index.pop((instance.environment_id, instance.image_id, instance.name), None)
        self.instance_index[(instance.environment_id, instance.image_id)].discard(instance_id)
        self.environment_instance[instance.environment_id].discard(instance_id)
        self.image_instance[instance.image_id].discard(instance_id)
        return instance


class LocalMetadataRepository(MetadataRepository):
    """
    :class:`.MetadataRepository` implementation which stores metadata in a local filesystem as JSON file.

    Warning: file storage is completely overwritten on each update,
    thus this repository is not suitable for high-performance scenarios.

    :param path: path to json with the metadata, if `None` metadata is stored in-memory.
    """

    type = 'local'

    def __init__(self, path=None):
        self.path = path
        if self.path is not None:
            os.makedirs(os.path.dirname(self.path), exist_ok=True)

        self.data: _LocalContainer = _LocalContainer()
        self.load()
        self.save()

    def load(self):
        if self.path is not None and os.path.exists(self.path):
            with open(self.path, 'r', encoding='utf8') as f:
                logger.debug('Loading metadata from %s', self.path)
                self.data = pyjackson.load(f, _LocalContainer)
        else:
            self.data = _LocalContainer()

    def save(self):
        if self.path is None:
            return
        with open(self.path, 'w', encoding='utf8') as f:
            logger.debug('Saving metadata to %s', self.path)
            pyjackson.dump(f, self.data)

    @bind_to_self
    def get_projects(self) -> List[Project]:
        return copy.deepcopy([self.data.get_project_by_id(p) for p in self.data.projects.keys()])

    @bind_to_self
    def get_project_by_name(self, name: str) -> Project:
        return copy.deepcopy(self.data.get_project_by_name(name))

    @bind_to_self
    def get_project_by_id(self, id) -> Project:
        return copy.deepcopy(self.data.get_project_by_id(id))

    @bind_to_self
    def create_project(self, project: Project) -> Project:
        if self.get_project_by_name(project.name) is not None:
            raise ExistingProjectError(project)
        project._id = self.data.get_and_increment('next_project_id')
        self.data.add_project(copy.deepcopy(project))
        self.save()
        return project
    #get_or_create_submodel

    def update_project(self, project: Project) -> Project:
        existing_project = self.get_project_by_id(project.id)
        if existing_project is None:
            raise NonExistingProjectError(project)

        self.data.remove_project(project.id)
        proj_copy = copy.deepcopy(project)
        self.data.add_project(proj_copy)
        for task in proj_copy.tasks.values():
            self.save_task(task)
        self.save()
        return project

    def delete_project(self, project: Project):
        try:
            if self.get_tasks(project):
                raise ProjectWithTasksError(project)
            self.data.remove_project(project.id)
            self.save()
            project.unbind_meta_repo()
        except (KeyError, AttributeError):
            raise NonExistingProjectError(project)

    @bind_to_self
    def get_tasks(self, project: ProjectVar) -> List[Task]:
        project = self._resolve_project(project)
        return copy.deepcopy(list(project.tasks.values()))

    @bind_to_self
    def get_task_by_name(self, project: ProjectVar, task_name: str) -> Optional[Task]:
        project = self._resolve_project(project)
        if project is None:
            return None
        return copy.deepcopy(self.data.get_task_by_name(project.id, task_name))

    @bind_to_self
    def get_task_by_id(self, id) -> Task:
        return copy.deepcopy(self.data.get_task_by_id(id))

    @bind_to_self
    def create_task(self, task: Task) -> Task:
        self._validate_task(task)

        existing_project = self.get_project_by_id(task.project_id)
        if existing_project is None:
            raise NonExistingProjectError(task.project_id)

        existing_task = self.get_task_by_name(existing_project, task.name)
        if existing_task is not None:
            raise ExistingTaskError(task)

        task._id = self.data.get_and_increment('next_task_id')
        self.data.add_task(copy.deepcopy(task))
        self.save()
        return task

    def update_task(self, task: Task) -> Task:
        if task.id is None or self.get_task_by_id(task.id) is None:
            raise NonExistingTaskError(task)
        self._validate_task(task)

        existing_project = self.get_project_by_id(task.project_id)
        if existing_project is None:
            raise NonExistingProjectError(task.project_id)

        self.data.remove_task(task.id)
        task_copy = copy.deepcopy(task)
        self.data.add_task(task_copy)
        for model in task_copy.models.values():
            self.save_model(model)
        for pipeline in task_copy.pipelines.values():
            self.save_pipeline(pipeline)
        for image in task_copy.images.values():
            self.save_image(image)
        self.save()
        return task

    def delete_task(self, task: Task):
        if task.id is None:
            raise NonExistingTaskError(task)
        if self.get_models(task) or self.get_pipelines(task) or self.get_images(task):
            raise TaskWithFKError(task)
        self.data.remove_task(task.id)
        self.save()
        task.unbind_meta_repo()

    @bind_to_self
    def get_models(self, task: TaskVar, project: ProjectVar = None) -> List[Model]:
        task = self._resolve_task(task, project)
        return copy.deepcopy(list(task.models.values()))

    @bind_to_self
    def get_model_by_name(self, model_name: str, task: TaskVar, project: ProjectVar = None) -> Optional[Model]:
        task = self._resolve_task(task, project)
        if task is None:
            return None
        return copy.deepcopy(self.data.get_model_by_name(task.id, model_name))

    @bind_to_self
    def get_model_by_id(self, id) -> Model:
        return copy.deepcopy(self.data.get_model_by_id(id))

    @bind_to_self
    def create_model(self, model: Model) -> Model:
        self._validate_model(model)

        existing_task = self.get_task_by_id(model.task_id)
        if existing_task is None:
            raise NonExistingTaskError(model.task_id)

        if self.get_model_by_name(model.name, existing_task) is not None:
            raise ExistingModelError(model)

        model._id = self.data.get_and_increment('next_model_id')
        self.data.add_model(copy.deepcopy(model))
        self.save()
        return model

    def update_model(self, model: Model) -> Model:
        self._validate_model(model)

        task = self.get_task_by_id(model.task_id)
        if task is None:
            raise NonExistingTaskError(model.task_id)

        existing_model = self.get_model_by_id(model.id)
        if existing_model is None:
            raise NonExistingModelError(model)

        self.data.remove_model(model.id)
        model_copy = copy.deepcopy(model)
        self.data.add_model(model_copy)
        self.save()
        return model

    def delete_model(self, model: Model):
        if model.id is None:
            raise NonExistingModelError(model)
        self.data.remove_model(model.id)
        self.save()
        model.unbind_meta_repo()

    # ____________-

    @bind_to_self
    def get_pipelines(self, task: TaskVar, project: ProjectVar = None) -> List[Pipeline]:
        task = self._resolve_task(task, project)
        return copy.deepcopy(list(task.pipelines.values()))

    @bind_to_self
    def get_pipeline_by_name(self, pipeline_name: str, task: TaskVar, project: ProjectVar = None) -> Optional[Pipeline]:
        task = self._resolve_task(task, project)
        if task is None:
            return None
        return copy.deepcopy(self.data.get_pipeline_by_name(task.id, pipeline_name))

    @bind_to_self
    def get_pipeline_by_id(self, id) -> Pipeline:
        return copy.deepcopy(self.data.get_pipeline_by_id(id))

    @bind_to_self
    def create_pipeline(self, pipeline: Pipeline) -> Pipeline:
        self._validate_pipeline(pipeline)

        existing_task = self.get_task_by_id(pipeline.task_id)
        if existing_task is None:
            raise NonExistingTaskError(pipeline.task_id)

        if self.get_pipeline_by_name(pipeline.name, existing_task) is not None:
            raise ExistingPipelineError(pipeline)

        pipeline._id = self.data.get_and_increment('next_pipeline_id')
        self.data.add_pipeline(copy.deepcopy(pipeline))
        self.save()
        return pipeline

    def update_pipeline(self, pipeline: Pipeline) -> Pipeline:
        self._validate_pipeline(pipeline)

        task = self.get_task_by_id(pipeline.task_id)
        if task is None:
            raise NonExistingTaskError(pipeline.task_id)

        existing_pipeline = self.get_pipeline_by_id(pipeline.id)
        if existing_pipeline is None:
            raise NonExistingPipelineError(pipeline)

        self.data.remove_pipeline(pipeline.id)
        pipeline_copy = copy.deepcopy(pipeline)
        self.data.add_pipeline(pipeline_copy)
        self.save()
        return pipeline

    def delete_pipeline(self, pipeline: Pipeline):
        if pipeline.id is None:
            raise NonExistingPipelineError(pipeline)
        self.data.remove_pipeline(pipeline.id)
        self.save()
        pipeline.unbind_meta_repo()

    # _____________

    @bind_to_self
    def get_images(self, task: TaskVar, project: ProjectVar = None) -> List[Image]:
        task = self._resolve_task(task, project)
        return copy.deepcopy(list(task.images.values()))

    @bind_to_self
    def get_image_by_name(self, image_name, task: TaskVar, project: ProjectVar = None) -> Optional[Image]:
        task = self._resolve_task(task, project)
        return copy.deepcopy(self.data.get_image_by_name(task.id, image_name))

    @bind_to_self
    def get_image_by_id(self, id: int) -> Optional[Image]:
        return copy.deepcopy(self.data.get_image_by_id(id))

    @bind_to_self
    def create_image(self, image: Image) -> Image:
        self._validate_image(image)

        task = self.get_task_by_id(image.task_id)
        if task is None:
            raise NonExistingTaskError(image.task_id)

        if self.get_image_by_name(image.name, task) is not None:
            raise ExistingImageError(image)

        image._id = self.data.get_and_increment('next_image_id')
        self.data.add_image(copy.deepcopy(image))
        self.save()
        return image

    def update_image(self, image: Image) -> Image:
        self._validate_image(image)

        existing_task = self.get_task_by_id(image.task_id)
        if existing_task is None:
            raise NonExistingTaskError(image.task_id)

        existing_image = self.get_image_by_id(image.id)
        if existing_image is None:
            raise NonExistingImageError(image)

        self.data.remove_image(image.id)
        self.data.add_image(copy.deepcopy(image))
        self.save()
        return image

    def delete_image(self, image: Image):
        if self.data.get_instances_by_image_id(image.id):
            raise ImageWithInstancesError(image)
        if image.id is None:
            raise NonExistingImageError(image)
        self.data.remove_image(image.id)
        self.save()
        image.unbind_meta_repo()

    @bind_to_self
    def get_environments(self) -> List[RuntimeEnvironment]:
        return copy.deepcopy([self.data.get_environment_by_id(e) for e in self.data.environments.keys()])

    @bind_to_self
    def get_environment_by_name(self, name) -> Optional[RuntimeEnvironment]:
        return copy.deepcopy(self.data.get_environment_by_name(name))

    @bind_to_self
    def get_environment_by_id(self, id: int) -> Optional[RuntimeEnvironment]:
        return copy.deepcopy(self.data.get_environment_by_id(id))

    @bind_to_self
    def create_environment(self, environment: RuntimeEnvironment) -> RuntimeEnvironment:
        self._validate_environment(environment)

        if self.get_environment_by_name(environment.name) is not None:
            raise ExistingEnvironmentError(environment)
        environment._id = self.data.get_and_increment('next_environment_id')
        self.data.add_environment(copy.deepcopy(environment))
        self.save()
        return environment

    def update_environment(self, environment: RuntimeEnvironment) -> RuntimeEnvironment:
        self._validate_environment(environment)

        existing_environment = self.get_environment_by_id(environment.id)
        if existing_environment is None:
            raise NonExistingEnvironmentError(environment)

        self.data.remove_environment(environment.id)
        self.data.add_environment(copy.deepcopy(environment))
        self.save()
        return environment

    def delete_environment(self, environment: RuntimeEnvironment):
        if self.data.get_instances_by_environment_id(environment.id):
            raise EnvironmentWithInstancesError(environment)
        try:
            self.data.remove_environment(environment.id)
            self.save()
            environment.unbind_meta_repo()
        except (KeyError, AttributeError):
            raise NonExistingEnvironmentError(environment)

    @bind_to_self
    def get_instances(self, image: Union[int, Image] = None, environment: Union[int, RuntimeEnvironment] = None) \
            -> List[RuntimeInstance]:
        if image is None and environment is None:
            raise ValueError('Image and environment were not provided to the function')
        if image is not None:
            image = image.id if isinstance(image, Image) else image
        if environment is not None:
            environment = environment.id if isinstance(environment, RuntimeEnvironment) else environment

        if image is not None and environment is not None:
            return self.data.get_instances(environment, image)
        elif image is not None:
            return self.data.get_instances_by_image_id(image)
        else:
            return self.data.get_instances_by_environment_id(environment)

    @bind_to_self
    def get_instance_by_name(self, instance_name, image: Union[int, Image],
                             environment: Union[int, RuntimeEnvironment]) -> Optional[RuntimeInstance]:
        image = image.id if isinstance(image, Image) else image
        environment = environment.id if isinstance(environment, RuntimeEnvironment) else environment
        return self.data.get_instance_by_name(environment, image, instance_name)

    @bind_to_self
    def get_instance_by_id(self, id: int) -> Optional[RuntimeInstance]:
        return self.data.get_instance_by_id(id)

    @bind_to_self
    def create_instance(self, instance: RuntimeInstance) -> RuntimeInstance:
        self._validate_instance(instance)

        image = self.get_image_by_id(instance.image_id)
        if image is None:
            raise NonExistingImageError(instance.image_id)

        environment = self.get_environment_by_id(instance.environment_id)
        if environment is None:
            raise NonExistingEnvironmentError(instance.environment_id)

        if self.get_instance_by_name(instance.name, image, environment) is not None:
            raise ExistingInstanceError(instance)

        instance._id = self.data.get_and_increment('next_instance_id')
        self.data.add_instance(copy.deepcopy(instance))
        self.save()
        return instance

    def update_instance(self, instance: RuntimeInstance) -> RuntimeInstance:
        self._validate_instance(instance)

        existing_instance = self.get_instance_by_id(instance.id)
        if existing_instance is None:
            raise NonExistingInstanceError(instance)

        self.data.remove_instance(instance.id)
        self.data.add_instance(copy.deepcopy(instance))
        self.save()
        return instance

    def delete_instance(self, instance: RuntimeInstance):
        try:
            self.data.remove_instance(instance.id)
            self.save()
            instance.unbind_meta_repo()
        except (KeyError, AttributeError):
            raise NonExistingInstanceError(instance)



import inspect
import re
from abc import abstractmethod
from typing import Optional

from pyjackson.utils import get_function_signature

EXPOSED_FIELD = '_client_exposed'


class ExposedMethod:
    def __init__(self, name: str = None):
        self.name = name
        self.method = None

    def __call__(self, method):
        setattr(method, EXPOSED_FIELD, self)
        self.method = method
        if self.name is None:
            self.name = self.original_name
        return method

    def __repr__(self):
        return f'{self.name}_method'

    @property
    def original_name(self):
        return self.method.__name__

    @abstractmethod
    def generate_code(self):
        """Generate method code"""
        raise NotImplementedError()

    def get_declaration(self):
        dec = ''
        for line in inspect.getsourcelines(self.method)[0]:
            strip = line.strip()
            if not strip or (dec == '' and not strip.startswith('def')):
                continue
            dec += line
            if re.sub(re.compile("#.*?\n"), "", line).strip().endswith(':'):
                break
        return dec.strip().replace(self.original_name, self.name)

    def get_signature(self):
        return get_function_signature(self.method)


def get_exposed_method(f) -> Optional[ExposedMethod]:
    return getattr(f, EXPOSED_FIELD, None)



import contextlib
import datetime
from typing import List, Optional, Type, TypeVar, Union

from sqlalchemy import create_engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session, sessionmaker

from algolink.core.errors import (EnvironmentWithInstancesError, ExistingEnvironmentError, ExistingImageError,
                                 ExistingInstanceError, ExistingModelError, ExistingPipelineError, ExistingProjectError,
                                 ExistingTaskError, ImageWithInstancesError, NonExistingEnvironmentError,
                                 NonExistingImageError, NonExistingInstanceError, NonExistingModelError,
                                 NonExistingPipelineError, NonExistingProjectError, NonExistingTaskError,
                                 ProjectWithTasksError, TaskWithFKError, UnknownMetadataError, ExistingExperimentError,
                                  NonExistingExperimentError,ExistingModelMetricError)
from algolink.core.objects.core import (AlgoLinkObject, Image, Model, Pipeline, Project, RuntimeEnvironment,
                                       RuntimeInstance, Task, Experiment, ModelMetric, ModelParam, BestResult)
from algolink.repository.metadata import MetadataRepository
from algolink.repository.metadata.base import ProjectVar, TaskVar, bind_to_self
from algolink.utils.log import logger

from .models import (Attaching, Base, SImage, SModel, SPipeline, SProject, SRuntimeEnvironment, SRuntimeInstance, STask,
                     update_attrs, TExperiment, TModelmetric, TModelparam, TBestresult)

T = TypeVar('T', bound=AlgoLinkObject)


class SQLAlchemyMetaRepository(MetadataRepository):
    """
    :class:`.MetadataRepository` implementation which stores metadata in SQL database via `sqlalchemy` library.

    :param db_uri: URI of SQL database to connect to
    """

    type = 'sqlalchemy'
    experiments: Type[TExperiment] = TExperiment
    #submodels: Type[TSubmodel] = TSubmodel
    model_metrics: Type[TModelmetric] = TModelmetric
    model_params: Type[TModelparam] = TModelparam
    best_result: Type[TBestresult] = TBestresult
    projects: Type[SProject] = SProject
    tasks: Type[STask] = STask
    models: Type[SModel] = SModel
    pipelines: Type[SPipeline] = SPipeline
    images: Type[SImage] = SImage
    environments: Type[SRuntimeEnvironment] = SRuntimeEnvironment
    instances: Type[SRuntimeInstance] = SRuntimeInstance

    def __init__(self, db_uri: str):
        self.db_uri = db_uri
        self._engine = create_engine(db_uri)
        Base.metadata.create_all(self._engine)
        self._Session = sessionmaker(bind=self._engine)
        self._active_session = None

    @contextlib.contextmanager
    def _session(self) -> Session:
        if self._active_session is None:
            logger.debug('Creating session for %s', self.db_uri)
            self._active_session = self._Session()
            new_session = True
        else:
            new_session = False

        try:
            yield self._active_session

            if new_session:
                self._active_session.commit()
        except:  # noqa
            if new_session:
                self._active_session.rollback()
            raise
        finally:
            if new_session:
                self._active_session.close()
                self._active_session = None

    def _get_objects(self, object_type: Type[Attaching], add_filter=None) -> List:
        with self._session() as s:
            if add_filter is None:
                logger.debug('Getting %ss', object_type.__name__)
            else:
                logger.debug('Getting %ss with filter %s', object_type.__name__, add_filter)
            q = s.query(object_type)
            if add_filter is not None:
                q = q.filter(add_filter)
            return [o.to_obj() for o in q.all()]

    def _get_object_by_name(self, object_type: Type[Attaching], name, add_filter=None):
        with self._session() as s:
            if add_filter is None:
                logger.debug('Getting %s with name %s', object_type.__name__, name)
            else:
                logger.debug('Getting %s with name %s with filter %s', object_type.__name__, name, add_filter)
            q = s.query(object_type).filter(object_type.name == name)
            if add_filter is not None:
                q = q.filter(add_filter)
            obj = q.first()
            if obj is None:
                return
            return obj.to_obj()

    def _get_sql_object_by_id(self, object_type: Type[Attaching], id: int):
        with self._session() as s:
            logger.debug('Getting %s[%s]', object_type.__name__, id)
            obj = s.query(object_type).filter(object_type.id == id).first()
            if obj is None:
                return
            return obj

    def _get_object_by_id(self, object_type: Type[Attaching], id: int):
        with self._session():
            sql_obj = self._get_sql_object_by_id(object_type, id)
            return sql_obj.to_obj() if sql_obj is not None else None

    def _create_object(self, object_type: Type[Attaching], obj: T, error_type) -> T:
        with self._session() as s:
            p = object_type.from_obj(obj, new=True)
            s.add(p)
            try:
                logger.debug('Inserting object %s', p)
                s.commit()
            except IntegrityError:
                raise error_type(obj)
            obj._id = p.id
            return obj

    def _delete_object(self, object_type: Type[Attaching], obj, ne_error_type, ie_error_type):
        with self._session() as s:
            p = s.query(object_type).filter(object_type.id == obj.id).first()
            if p is None:
                raise ne_error_type(obj)
            logger.debug('Deleting object %s', p)
            try:
                s.delete(p)
                s.commit()
            except IntegrityError:
                s.rollback()
                if p.to_obj().bind_meta_repo(self).has_children():
                    raise ie_error_type(obj)
                else:
                    raise UnknownMetadataError

    @bind_to_self
    def get_projects(self) -> List[Project]:
        return self._get_objects(self.projects)

    @bind_to_self
    def get_project_by_name(self, name: str) -> Optional[Project]:
        return self._get_object_by_name(self.projects, name)

    @bind_to_self
    def get_project_by_id(self, id: int) -> Optional[Project]:
        return self._get_object_by_id(self.projects, id)

    @bind_to_self
    def create_project(self, project: Project) -> Project:
        self._validate_project(project)
        return self._create_object(self.projects, project, ExistingProjectError)

    def update_project(self, project: Project) -> Project:
        with self._session():
            p: SProject = self._get_sql_object_by_id(self.projects, project.id)
            if p is None:
                raise NonExistingProjectError(project)
            kwargs = SProject.get_kwargs(project)
            kwargs.pop('tasks', None)
            update_attrs(p, **kwargs)

            for t in p.tasks:
                if t.id in project.tasks:
                    self.update_task(project.tasks.get(t.id))
                else:
                    project._tasks.add(t.to_obj())
            return project

    def delete_project(self, project: Project):
        self._delete_object(self.projects, project, NonExistingProjectError, ProjectWithTasksError)
        project.unbind_meta_repo()

    @bind_to_self
    def get_tasks(self, project: ProjectVar) -> List[Task]:
        project = self._resolve_project(project)
        return self._get_objects(self.tasks, self.tasks.project_id == project.id)

    @bind_to_self
    def get_task_by_name(self, project: ProjectVar, task_name: str) -> Optional[Task]:
        p = self._resolve_project(project)
        if p is None:
            return None
        return self._get_object_by_name(self.tasks, task_name, self.tasks.project_id == p.id)

    @bind_to_self
    def get_task_by_id(self, id: int) -> Optional[Task]:
        return self._get_object_by_id(self.tasks, id)

    @bind_to_self
    def create_task(self, task: Task) -> Task:
        self._validate_task(task)
        if self._get_sql_object_by_id(self.projects, task.project_id) is None:
            raise NonExistingProjectError(task.project_id)
        return self._create_object(self.tasks, task, ExistingTaskError)

    def update_task(self, task: Task) -> Task:
        with self._session():
            t: STask = self._get_sql_object_by_id(self.tasks, task.id)
            if t is None:
                raise NonExistingTaskError(task)
            kwargs = STask.get_kwargs(task)
            kwargs.pop('models', None)
            kwargs.pop('images', None)
            kwargs.pop('pipelines', None)
            update_attrs(t, **kwargs)

            for m in t.models:
                if m.id in task.models:
                    self.update_model(task.models.get(m.id))
                else:
                    task._models.add(m.to_obj())
            for p in t.pipelines:
                if p.id in task.pipelines:
                    self.update_pipeline(task.pipelines.get(p.id))
                else:
                    task._images.add(p.to_obj())

            for i in t.images:
                if i.id in task.images:
                    self.update_image(task.images.get(i.id))
                else:
                    task._images.add(i.to_obj())
            return task

    def delete_task(self, task: Task):
        self._delete_object(self.tasks, task, NonExistingTaskError, TaskWithFKError)
        task.unbind_meta_repo()

    @bind_to_self
    def get_models(self, task: TaskVar, project: ProjectVar = None) -> List[Model]:
        task = self._resolve_task(task, project)
        return self._get_objects(self.models, self.models.task_id == task.id)

    @bind_to_self
    def get_model_by_name(self, model_name, task: TaskVar, project: ProjectVar = None) -> Optional[Model]:
        task = self._resolve_task(task, project)
        if task is None:
            return None
        return self._get_object_by_name(self.models, model_name, self.models.task_id == task.id)

    @bind_to_self
    def get_model_by_id(self, id: int) -> Optional[Model]:
        return self._get_object_by_id(self.models, id)

    @bind_to_self
    def create_model(self, model: Model) -> Model:
        self._validate_model(model)
        if self._get_sql_object_by_id(self.tasks, model.task_id) is None:
            raise NonExistingTaskError(model.task_id)
        return self._create_object(self.models, model, ExistingModelError)

    def update_model(self, model: Model) -> Model:
        with self._session():
            m: SModel = self._get_sql_object_by_id(self.models, model.id)
            if m is None:
                raise NonExistingModelError(model)
            kwargs = SModel.get_kwargs(model)
            update_attrs(m, **kwargs)

            return model

    def delete_model(self, model: Model):
        self._delete_object(self.models, model, NonExistingModelError, AssertionError)
        model.unbind_meta_repo()

    @bind_to_self
    def get_pipelines(self, task: TaskVar, project: ProjectVar = None) -> List[Pipeline]:
        task = self._resolve_task(task, project)
        return self._get_objects(self.pipelines, self.pipelines.task_id == task.id)

    @bind_to_self
    def get_pipeline_by_name(self, pipeline_name, task: TaskVar, project: ProjectVar = None) -> Optional[Pipeline]:
        task = self._resolve_task(task, project)
        if task is None:
            return None
        return self._get_object_by_name(self.pipelines, pipeline_name, self.pipelines.task_id == task.id)

    @bind_to_self
    def get_pipeline_by_id(self, id: int) -> Optional[Pipeline]:
        return self._get_object_by_id(self.pipelines, id)

    @bind_to_self
    def create_pipeline(self, pipeline: Pipeline) -> Pipeline:
        self._validate_pipeline(pipeline)
        if self._get_sql_object_by_id(self.tasks, pipeline.task_id) is None:
            raise NonExistingTaskError(pipeline.task_id)
        return self._create_object(self.pipelines, pipeline, ExistingPipelineError)

    def update_pipeline(self, pipeline: Pipeline) -> Pipeline:
        with self._session():
            m: SPipeline = self._get_sql_object_by_id(self.pipelines, pipeline.id)
            if m is None:
                raise NonExistingPipelineError(pipeline)
            kwargs = SPipeline.get_kwargs(pipeline)
            update_attrs(m, **kwargs)

            return pipeline

    def delete_pipeline(self, pipeline: Pipeline):
        self._delete_object(self.pipelines, pipeline, NonExistingPipelineError, AssertionError)
        pipeline.unbind_meta_repo()

    @bind_to_self
    def get_images(self, task: TaskVar, project: ProjectVar = None) -> List[Image]:
        task = self._resolve_task(task, project)
        return self._get_objects(self.images, self.images.task_id == task.id)

    @bind_to_self
    def get_image_by_name(self, image_name, task: TaskVar, project: ProjectVar = None) -> Optional[Image]:
        task = self._resolve_task(task, project)
        if task is None:
            return None
        return self._get_object_by_name(self.images, image_name, self.images.task_id == task.id)

    @bind_to_self
    def get_image_by_id(self, id: int) -> Optional[Image]:
        return self._get_object_by_id(self.images, id)

    @bind_to_self
    def create_image(self, image: Image) -> Image:
        self._validate_image(image)
        if self._get_sql_object_by_id(self.tasks, image.task_id) is None:
            raise NonExistingTaskError(image.task_id)
        return self._create_object(self.images, image, ExistingImageError)

    def update_image(self, image: Image) -> Image:
        with self._session():
            i = self._get_sql_object_by_id(self.images, image.id)
            if i is None:
                raise NonExistingImageError(image)
            update_attrs(i, **SImage.get_kwargs(image))
            return image

    def delete_image(self, image: Image):
        self._delete_object(self.images, image, NonExistingImageError, ImageWithInstancesError)
        image.unbind_meta_repo()

    @bind_to_self
    def get_environments(self) -> List[RuntimeEnvironment]:
        return self._get_objects(self.environments)

    @bind_to_self
    def get_environment_by_name(self, name) -> Optional[RuntimeEnvironment]:
        return self._get_object_by_name(self.environments, name)

    @bind_to_self
    def get_environment_by_id(self, id: int) -> Optional[RuntimeEnvironment]:
        return self._get_object_by_id(self.environments, id)

    @bind_to_self
    def create_environment(self, environment: RuntimeEnvironment) -> RuntimeEnvironment:
        self._validate_environment(environment)
        return self._create_object(self.environments, environment, ExistingEnvironmentError)

    def update_environment(self, environment: RuntimeEnvironment) -> RuntimeEnvironment:
        with self._session():
            i = self._get_sql_object_by_id(self.environments, environment.id)
            if i is None:
                raise NonExistingEnvironmentError(environment)
            update_attrs(i, **SRuntimeEnvironment.get_kwargs(environment))
            return environment

    def delete_environment(self, environment: RuntimeEnvironment):
        self._delete_object(self.environments, environment, NonExistingEnvironmentError, EnvironmentWithInstancesError)
        environment.unbind_meta_repo()

    @bind_to_self
    def get_instances(self, image: Union[int, Image] = None, environment: Union[int, RuntimeEnvironment] = None) \
            -> List[RuntimeInstance]:
        if image is None and environment is None:
            raise ValueError('Image and environment were not provided to the function')

        filter = True
        if image is not None:
            image = image.id if isinstance(image, Image) else image
            filter = filter and self.instances.image_id == image
        if environment is not None:
            environment = environment.id if isinstance(environment, RuntimeEnvironment) else environment
            filter = filter and self.instances.environment_id == environment

        return self._get_objects(self.instances, filter)

    @bind_to_self
    def get_instance_by_name(self, instance_name, image: Union[int, Image],
                             environment: Union[int, RuntimeEnvironment]) -> Optional[RuntimeInstance]:
        image = image.id if isinstance(image, Image) else image
        environment = environment.id if isinstance(environment, RuntimeEnvironment) else environment
        return self._get_object_by_name(self.instances, instance_name,
                                        self.instances.image_id == image and
                                        self.instances.environment_id == environment)

    @bind_to_self
    def get_instance_by_id(self, id: int) -> Optional[RuntimeInstance]:
        return self._get_object_by_id(self.instances, id)

    @bind_to_self
    def create_instance(self, instance: RuntimeInstance) -> RuntimeInstance:
        self._validate_instance(instance)
        if self._get_sql_object_by_id(self.images, instance.image_id) is None:
            raise NonExistingImageError(instance.image_id)
        return self._create_object(self.instances, instance, ExistingInstanceError)

    def update_instance(self, instance: RuntimeInstance) -> RuntimeInstance:
        with self._session():
            i = self._get_sql_object_by_id(self.instances, instance.id)
            if i is None:
                raise NonExistingInstanceError(instance)
            update_attrs(i, **SRuntimeInstance.get_kwargs(instance))
            return instance

    def delete_instance(self, instance: RuntimeInstance):
        self._delete_object(self.instances, instance, NonExistingInstanceError, AssertionError)
        instance.unbind_meta_repo()
        
    ## tracking ml -leepand
    @bind_to_self
    def get_experiments(self):
        return self._get_objects(self.experiments)

    @bind_to_self
    def get_experiment_by_name(self, task: str, name: str)-> Optional[Experiment]:
        return self._get_object_by_name(self.experiments, name)

    # get submodels count/task_id
    @bind_to_self
    def get_experiment_sequence(self,id: str)-> Optional[Experiment]:
        with self._session() as s:
            object_type = self.experiments
            logger.debug('Getting %s[%s]', object_type.__name__, id)
            obj_cnt = s.query(object_type).filter(object_type.task_id == id).count()
            if obj_cnt is None:
                return
            return obj_cnt
        #return self._get_object_by_id(self.experiments, id)

    @bind_to_self
    def get_experiment_by_id(self, id: int):
        return self._get_object_by_id(self.experiments, id)
    @bind_to_self
    def create_experiment(self, experiment: Experiment) -> Experiment:
        self._validate_experiment(experiment)
        #print(exp_name,experiment.name,"exp_name")
        exp_name = self.get_experiment_by_name(experiment.task_id,experiment.name)
        
        if exp_name is not None:
            raise ExistingExperimentError(exp_name)
        if self._get_sql_object_by_id(self.tasks, experiment.task_id) is None:
            raise NonExistingExperimentError(experiment.task_id)
        return self._create_object(self.experiments, experiment, ExistingExperimentError)

    @bind_to_self
    def get_metric_by_name(self, experiment_id: int, name: str)-> Optional[ModelMetric]:
        return self._get_object_by_name(self.model_metrics, name, self.model_metrics.experiment_id == experiment_id)
    
    @bind_to_self
    def get_metric_by_name_epoch(self, experiment_id: int, name: str,epoch: int)-> Optional[ModelMetric]:
        filters = {
            #self.model_metrics.name = name,
            self.model_metrics.experiment_id == experiment_id,
            self.model_metrics.epoch== epoch
        }
        with self._session() as s:
            q = s.query(self.model_metrics).filter(self.model_metrics.name == name)
            add_filter = 1
            if add_filter is not None:
                q = q.filter(*filters)
            obj = q.first()
            if obj is None:
                return
            return obj.to_obj()
        #return self._get_object_by_name(self.model_metrics, name, 
        #                                (self.model_metrics.experiment_id == experiment_id , 
        #                                self.model_metrics.epoch== epoch))
    
    @bind_to_self
    def create_metric(self, metric: ModelMetric) -> ModelMetric:
        return self._create_object(self.model_metrics, metric, ExistingModelMetricError)
        
        
    @bind_to_self
    def create_param(self, param: ModelParam) -> ModelParam:
        return self._create_object(self.model_params, param, ExistingModelMetricError)

    @bind_to_self
    def create_best_result(self, best_result: BestResult) -> BestResult:
        return self._create_object(self.best_result, best_result, ExistingModelMetricError)

    def update_experiment(self, id):
        with self._session():
            e: TExperiment = self._get_sql_object_by_id(self.experiments, id)
            if e is None:
                raise NonExistingProjectError(experiment)
            kwargs = TExperiment.get_kwargs(experiment)
            kwargs.pop('submodels', None)
            update_attrs(p, **kwargs)

            for t in e.submodels:
                if t.id in experiment.submodels:
                    self.update_submodel(experiment.submodels.get(t.id))
                else:
                    experiment._submodels.add(t.to_obj())
            return experiment

    def delete_project(self, project: Project):
        self._delete_object(self.projects, project, NonExistingProjectError, ProjectWithTasksError)
        project.unbind_meta_repo()

#class Experiment()


from functools import wraps
from typing import List, Callable
from pyjackson.utils import get_function_fields, turn_args_to_kwargs


def validate_kwargs(f=None, *, allowed: List[str] = None, allowed_funcs: List[Callable] = None):

    def inner(func):
        all_fields = allowed or []
        if allowed_funcs is not None:
            all_fields += [field.name for func in allowed_funcs for field in get_function_fields(func, False)]
        all_fields = set(all_fields)
        fields = get_function_fields(func, False)
        all_fields.update(field.name for field in fields)

        @wraps(func)
        def inner_inner(self, *args, **kwargs):
            if len(args) > len(fields):
                raise TypeError(
                    f'{func.__name__}() takes {len(fields)} positional arguments but {len(args)} were given')
            kwargs = turn_args_to_kwargs(func, args, kwargs, False)
            extra_kwargs = set(kwargs.keys()).difference(all_fields)
            if len(extra_kwargs) > 0:
                raise TypeError(f'{extra_kwargs} are an invalid keyword arguments for this function')
            return func(self, **kwargs)

        return inner_inner

    if f is None:
        return inner
    return inner(f)



from .base import DockerContainer, DockerEnv, DockerImage, DockerIORegistry, DockerRegistry, RemoteRegistry
from .builder import DockerBuilder
from .helpers import build_docker_image, run_docker_instance
from .runner import DockerRunner, RunnerBase

__all__ = ['DockerRegistry', 'DockerContainer', 'DockerEnv', 'DockerImage', 'RemoteRegistry',
           'DockerRunner', 'RunnerBase', 'DockerBuilder',
           'DockerIORegistry', 'build_docker_image', 'run_docker_instance']



import typing

from algolink.core.errors import ArtifactExistsError, NoSuchArtifactError
from algolink.core.objects.artifacts import ArtifactCollection, Blob, Blobs, InMemoryBlob
from algolink.repository.artifact import ArtifactRepository


class InMemoryArtifactRepository(ArtifactRepository):
    """
    :class:`.ArtifactRepository` implementation which stores artifacts in-memory
    """

    type = 'inmemory'

    def __init__(self):
        self._cache: typing.Dict[str, ArtifactCollection] = {}

    def get_artifact(self, artifact_type, artifact_id: str) -> ArtifactCollection:
        artifact_id = f'{artifact_type}/{artifact_id}'
        if artifact_id not in self._cache:
            raise NoSuchArtifactError(artifact_id, self)

        return self._cache[artifact_id]

    def push_artifact(self, artifact_type, artifact_id: str, blobs: typing.Dict[str, Blob]) -> ArtifactCollection:
        artifact_id = f'{artifact_type}/{artifact_id}'
        if artifact_id in self._cache:
            raise ArtifactExistsError(artifact_id, self)
        self._cache[artifact_id] = Blobs({
            k: InMemoryBlob(v.bytes()) for k, v in blobs.items()
        })
        return self._cache[artifact_id]

    def delete_artifact(self, artifact_type, artifact_id: str):
        artifact_id = f'{artifact_type}/{artifact_id}'
        if artifact_id not in self._cache:
            raise NoSuchArtifactError(artifact_id, self)
        del self._cache[artifact_id]



import os
import time
from contextlib import contextmanager
from threading import Lock

import docker
import requests
from docker.errors import DockerException


def _is_docker_running(client: docker.DockerClient) -> bool:
    """
    Check if docker binary and docker daemon are available

    :param client: DockerClient instance
    :return: true or false
    """
    try:
        client.info()
        return True
    except (ImportError, IOError, DockerException):
        return False


def is_docker_running() -> bool:
    """
    Check if docker binary and docker daemon are available

    :return: true or false
    """
    with create_docker_client(check=False) as c:
        return _is_docker_running(c)


_docker_host_lock = Lock()


@contextmanager
def create_docker_client(docker_host: str = '', check=True) -> docker.DockerClient:
    """
    Context manager for DockerClient creation

    :param docker_host: DOCKER_HOST arg for DockerClient
    :param check: check if docker is available
    :return: DockerClient instance
    """
    with _docker_host_lock:
        os.environ["DOCKER_HOST"] = docker_host  # The env var DOCKER_HOST is used to configure docker.from_env()
        client = docker.from_env()
    if check and not _is_docker_running(client):
        raise RuntimeError("Docker daemon is unavailable")
    try:
        yield client
    finally:
        client.close()


def image_exists_at_dockerhub(tag):
    repo, tag = tag.split(':')
    resp = requests.get(f'https://registry.hub.docker.com/v1/repositories/{repo}/tags/{tag}')
    time.sleep(1)  # rate limiting
    return resp.status_code == 200


def repository_tags_at_dockerhub(repo):
    resp = requests.get(f'https://registry.hub.docker.com/v1/repositories/{repo}/tags')
    time.sleep(1)  # rate limiting
    if resp.status_code != 200:
        return {}
    else:
        return {tag['name'] for tag in resp.json()}



from . import config
from .client import AlgoLink, create_model
from .ext.ext_loader import ExtensionLoader, load_extensions
from .runtime.command_line import start_runtime

EBONITE_DEBUG = config.Core.DEBUG
if config.Core.AUTO_IMPORT_EXTENSIONS:
    ExtensionLoader.load_all()

__all__ = ['load_extensions', 'AlgoLink', 'ALGOLINK_DEBUG', 'start_runtime', 'create_model']
__version__ = '0.7.9'

if __name__ == '__main__':
    pass



import contextlib
import inspect
import os


def get_lib_path(*filename):
    path = __file__
    for _ in range(len(__name__.split('.')) - 1):
        path = os.path.dirname(path)
    return os.path.join(path, *filename)


def current_module_path(*path):
    stack = inspect.stack()
    caller_path = stack[1][1]
    return os.path.join(os.path.dirname(caller_path), *path)


@contextlib.contextmanager
def switch_curdir(path):
    """Context manager to temproary switch current dir"""
    prev_path = os.path.abspath('.')
    try:
        try:
            os.chdir(path)
        except FileNotFoundError:
            pass
        yield
    finally:
        os.chdir(prev_path)



from typing import Tuple, Type, Union

import numpy as np
from pyjackson.core import ArgList, Field
from pyjackson.errors import DeserializationError, SerializationError
from pyjackson.generics import Serializer

from algolink.core.analyzer.base import CanIsAMustHookMixin, TypeHookMixin
from algolink.core.analyzer.dataset import DatasetHook
from algolink.core.objects.dataset_type import DatasetType, LibDatasetTypeMixin
from algolink.core.objects.typing import ListTypeWithSpec, SizedTypedListType


def python_type_from_np_string_repr(string_repr: str) -> type:
    np_type = np_type_from_string(string_repr)
    return python_type_from_np_type(np_type)


def python_type_from_np_type(np_type: Union[Type, np.dtype]):
    if np_type.type == np.object_:
        return object
    value = np_type.type()
    if np_type.type.__module__ == 'numpy':
        value = value.item()
    return type(value)


def np_type_from_string(string_repr):
    try:
        return np.dtype(string_repr)
    except TypeError:
        raise ValueError('Unknown numpy type {}'.format(string_repr))


class NumpyNumberDatasetType(LibDatasetTypeMixin):
    """
    :class:`.DatasetType` implementation for `numpy.number` objects which
    converts them to built-in Python numbers and vice versa.

    :param dtype: `numpy.number` data type as string
    """
    libraries = [np]

    def __init__(self, dtype: str):
        self.dtype = dtype

    def get_spec(self) -> ArgList:
        return [Field(None, python_type_from_np_string_repr(self.dtype), False)]

    def deserialize(self, obj: dict) -> object:
        return self.actual_type(obj)

    def serialize(self, instance: np.number) -> object:
        self._check_type(instance, np.number, SerializationError)
        return instance.item()

    @property
    def actual_type(self):
        return np_type_from_string(self.dtype)

    def get_writer(self):
        from algolink.repository.dataset.artifact import PickleWriter
        return PickleWriter()


class NumpyNumberHook(CanIsAMustHookMixin, DatasetHook):
    """
    :class:`.DatasetHook` implementation for `numpy.number` objects which uses :class:`NumpyNumberDatasetType`.
    """

    def must_process(self, obj) -> bool:
        return isinstance(obj, np.number)

    def process(self, obj: np.number, **kwargs) -> DatasetType:
        return NumpyNumberDatasetType(obj.dtype.name)


class NumpyNdarrayHook(TypeHookMixin, DatasetHook):
    """
    :class:`.DatasetHook` implementation for `np.ndarray` objects which uses :class:`NumpyNdarrayDatasetType`
    """

    valid_types = [np.ndarray]

    def process(self, obj, **kwargs) -> DatasetType:
        return NumpyNdarrayDatasetType(obj.shape, obj.dtype.name)


class NumpyDTypeSerializer(Serializer):
    """
    PyJackson :class:`.Serializer` for `numpy` data types: stores types in JSON as their names.
    """

    def deserialize(self, obj: str):
        return getattr(np, obj)

    def serialize(self, instance) -> str:
        return str(instance)


class NumpyNdarrayDatasetType(ListTypeWithSpec, LibDatasetTypeMixin):
    """
    :class:`.DatasetType` implementation for `np.ndarray` objects
    which converts them to built-in Python lists and vice versa.

    :param shape: shape of `numpy.ndarray` objects in dataset
    :param dtype: data type of `numpy.ndarray` objects in dataset
    """

    real_type = np.ndarray
    libraries = [np]

    def __init__(self, shape: Tuple[int, ...], dtype: str):
        # TODO assert shape and dtypes len
        self.shape = (None,) + shape[1:]
        self.dtype = dtype

    def list_size(self):
        return self.shape[0]

    def _get_subtype(self, shape):
        if len(shape) == 0:
            return python_type_from_np_string_repr(self.dtype)
        elif len(shape) == 1:
            subtype = python_type_from_np_string_repr(self.dtype)
        else:
            subtype = self._get_subtype(shape[1:])
        return SizedTypedListType(shape[0], subtype)

    def get_spec(self) -> ArgList:
        return [Field(None, self._get_subtype(self.shape[1:]), False)]

    def deserialize(self, obj):
        try:
            ret = np.array(obj, dtype=np_type_from_string(self.dtype))
        except (ValueError, TypeError):
            raise DeserializationError(f'given object: {obj} could not be converted to array '
                                       f'of type: {np_type_from_string(self.dtype)}')
        self._check_shape(ret, DeserializationError)
        return ret

    def serialize(self, instance: np.ndarray):
        self._check_type(instance, np.ndarray, SerializationError)
        exp_type = np_type_from_string(self.dtype)
        if instance.dtype != exp_type:
            raise SerializationError(f'given array is of type: {instance.dtype}, expected: {exp_type}')
        self._check_shape(instance, SerializationError)
        return instance.tolist()

    def _check_shape(self, array, exc_type):
        if tuple(array.shape)[1:] != self.shape[1:]:
            raise exc_type(f'given array is of shape: {(None,) + tuple(array.shape)[1:]}, expected: {self.shape}')

    def get_writer(self):
        from .dataset_source import NumpyNdarrayWriter
        return NumpyNdarrayWriter()
